package org.ninjasoft.rovr.autopilots;
import org.ninjasoft.rovr.map.*;
import org.ninjasoft.rovr.net.*;
import java.util.*;

/**
 * This is the basic interface for creating your own autopilot class.  Extend this class and
 * then launch your extended class from the main UI window's code (RovrFrame).
 */
public abstract class BaseAutopilot implements MovementListener, Runnable {
	protected AutopilotStatus status;
	protected Client client;
	protected Mapper mapper;
	protected LinkedList responseQueue = new LinkedList();
	protected String moduleName = "Unknown";
	protected String moduleAuthor = "Unknown";
	
	public boolean running = true;
	
	private BaseAutopilot() {}
	
	public BaseAutopilot(AutopilotStatus status, Client client, Mapper mapper) {
		this.status = status;
		this.client = client;
		this.mapper = mapper;
		client.addMovementListener(this);
	}
	
	/**
	 * Helper function for calculating a right turn
	 * @param direction the current direction of travel
	 * @return the new direction, turning 90 degrees right
	 */
	public static int turnRight(int direction) {
		int result = direction;
		switch(direction) {
		case Client.COMMAND_UP: result = Client.COMMAND_RIGHT; break;
		case Client.COMMAND_RIGHT: result = Client.COMMAND_DOWN; break;
		case Client.COMMAND_DOWN: result = Client.COMMAND_LEFT; break;
		case Client.COMMAND_LEFT: result = Client.COMMAND_UP; break;
		}
		return result;
	}
	
	/**
	 * Helper function for calculating a left turn
	 * @param direction the current direction of travel
	 * @return the new direction, turning 90 degrees left
	 */
	public static int turnLeft(int direction) {
		int result = direction;
		switch(direction) {
		case Client.COMMAND_UP: result = Client.COMMAND_LEFT; break;
		case Client.COMMAND_LEFT: result = Client.COMMAND_DOWN; break;
		case Client.COMMAND_DOWN: result = Client.COMMAND_RIGHT; break;
		case Client.COMMAND_RIGHT: result = Client.COMMAND_UP; break;
		}
		return result;
	}

	/**
	 * Wait until we get a response code, timing out after the given number of seconds
	 * @param seconds
	 * @return
	 */
	protected int waitForCodeOrTimeout(int tenthsOfSeconds) {
		int max = tenthsOfSeconds;
		responseQueue.clear();
		for (int i=0; i<max; i++) {
			if (responseQueue.size() > 0) {
				int rc = ((Integer) responseQueue.removeFirst()).intValue();
				System.out.println("Queue returning " + rc);
				return rc;
			}
			try {Thread.sleep(100);} catch(Throwable t) {}
		}
		System.out.println("Queue timeout");
		return -1;
	}

	/**
	 * Main entry point.  Takes care of initialization and looping, chains to user functions.
	 */
	public void run() {
		int state = 0;
		System.out.println("Starting autopilot \"" + moduleName + "\" by " + moduleAuthor);
		try {
			initialize();
			while (running) {
				state = doState(state);
			}
			client.sendCommand(Client.COMMAND_STOP);
		}catch(Throwable t) {
			t.printStackTrace();
		}
		client.removeMovementListener(this);
		System.out.println("Autopilot \"" + moduleName + "\" has finished");
		status.autopilotFinished();
	}
	
	/**
	 * Given a current state, do something, then return the new state
	 * @param state
	 * @return
	 */
	abstract protected int doState(int state) throws Exception;
	
	/**
	 * Do whatever initialization you want
	 * @throws Exception
	 */
	abstract protected void initialize() throws Exception;

	public void movementResponse(int responseCode, String message) {
		System.out.println("Adding to queue" + responseCode);
		responseQueue.addLast(new Integer(responseCode));
		// If we've reached the end of the level, turn off autopilot
		if ((responseCode == 10) || (responseCode == 50))
			running = false;
	}
}
