package org.ninjasoft.rovr.net;
import java.net.*;
import java.io.*;
import java.util.Vector;
import java.util.Iterator;

/**
 * Client is the network library used by the system.  Pretty descriptive name, eh?
 * It handles all outbound and inbound communication.  You perform moves by
 * issuing sendCommand calls with the various COMMAND_* constants.  You get news of
 * the response events by subscribing to the listener queue with addMovementListener().
 * Everything else is pretty much just overhead. 
 *
 */
public class Client {
	public static final int CONTROL_PORT = 5399;
	public static final int RESPONSE_PORT = 5398;
	
	public static final int COMMAND_UP = 8;
	public static final int COMMAND_DOWN = 2;
	public static final int COMMAND_LEFT = 4;
	public static final int COMMAND_RIGHT = 6;
	public static final int COMMAND_STOP = 5;
	
	public static final int RESPONSE_MOVED = 20;
	public static final int RESPONSE_HIT_UP_DOWN = 30;
	public static final int RESPONSE_HIT_LEFT_RIGHT = 40;
	public static final int RESPONSE_MESSAGE = 50;
	public static final int RESPONSE_EXIT = 10;
	
	private int moveFactor = 15;
	private int leanDirection = 0;
    private int leanFactor = 0;

	private String ipAddress = "127.0.0.1";
	private Socket commandSocket = null;
	private Socket responseSocket = null;
	private InputStream in = null;
	private OutputStream out = null;
	private Vector listeners = new Vector();
	private Thread eventThread = null;
	
	public Client() {
	}
	
	public Client(String ipAddress) {
		this.ipAddress = ipAddress;
	}

	public void connect() throws UnknownHostException, IOException {
		this.commandSocket = new Socket(this.ipAddress, CONTROL_PORT);
		this.responseSocket = new Socket(this.ipAddress, RESPONSE_PORT);
		this.out = commandSocket.getOutputStream();
		this.in = responseSocket.getInputStream();
		this.eventThread = new Thread(new ResponseListener());
		this.eventThread.start();
	}
	
	public void setMoveFactor(int i) {
		System.out.println("Speed " + i);
		this.moveFactor = i;
	}
	
	public int getMoveFactor() {
		return this.moveFactor;
	}
	
	public void setLean(int direction, int factor) {
	    this.leanDirection = direction;
        this.leanFactor = factor;
	}
	
	public void sendCommand(int command) throws IOException {
		String commandString = "";
		int leanValue = this.leanFactor;
		if (this.leanDirection < 0)
			leanValue *= -1;
	    if (this.leanDirection == 0)
	       leanValue = 0;
		switch(command){
		case COMMAND_UP:
			commandString = "[" + leanValue + "," + (-1 * moveFactor) + "]\n";
			break;
		case COMMAND_DOWN:
			commandString = "[" + (-1 * leanValue) + "," + moveFactor + "]\n";
			break;
		case COMMAND_LEFT:
			commandString = "[" + (-1 * moveFactor) + "," + leanValue + "]\n";
			break;
		case COMMAND_RIGHT:
			commandString = "[" + moveFactor + "," + (-1 * leanValue) + "]\n";
			break;
		case COMMAND_STOP:
			commandString = "[0,0]\n";
			break;
		}
		if (commandString.length() > 0) {
			out.write(commandString.getBytes());
		}
	}
	
	public void addMovementListener(MovementListener listener) {
		this.listeners.add(listener);
	}
	
	public void removeMovementListener(MovementListener listener) {
		this.listeners.remove(listener);
	}
	
	private void notifyListeners(int code, String message) {
		for (Iterator i = this.listeners.iterator(); i.hasNext(); ) {
			((MovementListener) i.next()).movementResponse(code, message);
		}
	}
	
	public class ResponseListener implements Runnable {
		public void run() {
			try {
				while(true) {
					String message = "";
					// Read the code
					int ch = in.read();
					if (ch != -1) {
						int len = 0;
						// Check if there was an associated text string and read it
						len = (in.read() << 8) + in.read();
						for (int i=0; i<len; i++)
							message += (char) in.read();
						// Notify everyone registered as a listener
						notifyListeners(ch, message);
						// If we got a message, log it for posterity
						if (message.length() > 0) {
							FileOutputStream messageLog = new FileOutputStream("messages.txt", true);
							message += "\n";
							messageLog.write(message.getBytes());
							messageLog.close();
						}
					}
				}
			}catch(IOException e){
				System.out.println("Remote closed connection");
			}
		}
	}
}
