package org.ninjasoft.rovr.ui;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.EmptyBorder;
import java.util.Random;
import org.ninjasoft.rovr.net.*;
import org.ninjasoft.rovr.map.*;
import org.ninjasoft.rovr.autopilots.*;

/**
 * This is the main user interface for the ROVR controller.
 */
public class RovrFrame extends JFrame {
	/** UI and events initialized? */
    private boolean initialized = false;
    /** Inner class for handling actions */
    private Actions actions = new Actions();
    /** The network client, which actually performs the network talking and receiving of events */
    private Client client = null;
    /** An inner class to accept keyboard input */
    private RovrKeyAdapter rovrKeyAdapter = new RovrKeyAdapter();
    /** The mapping functionality.  This will end up being subscribed to the Client's event queue */
    private Mapper mapper = new Mapper();
    /** Panel which displays the map (the "view" to mapper's "model") */
    private MapFrame mapFrame = new MapFrame(mapper);
    /** Combo box for selecting speed multiplier */
    private JComboBox speedCombo = new JComboBox(new String[] {"15", "16", "17", "18", "19", "20"});
    /** Combo box for selecting a lean (works, but leaves mapper crazy) */
    private JComboBox leanSpeedCombo = new JComboBox(new String[] {"15", "16", "17", "18", "19", "20"});
    /** Combo box for selecting a lean (works, but leaves mapper crazy) */
    private JComboBox leanDirectionCombo = new JComboBox(new String[] {"left", "none", "right"});
    /** Text area for instructions */
	private JTextArea instructionText;
	/** Text area for status information */
	private JTextField statusText;
	/** Checkbox for whether or not to do the magnet thing */
	private JCheckBox magnetCheckbox = new JCheckBox("Magnet");
    /** Checkbox for autopilot */
    private JCheckBox autopilotCheckbox = new JCheckBox("Autopilot");
    /** The autopilot class */
    private BaseAutopilot autopilot = null;

	/** Instructions to display to the user */
    private static final String INSTRUCTIONS = 
    	"Movement is done using the keys 'A' and 'Z' on the left hand for up/down and " +
    	"',' and '.' on the right hand for left/right.  Arrow keys also work.  " +
    	"Use numbers 1 through 5 to adjust speed.  " + 
    	"Press [SPACE] to stop all movement.";

    /** Initialize GUI and events */
    public void initialize() {
        initializeGui();
        initializeEvents();
        this.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    }

    /** Initialize the GUI */
    private void initializeGui() {
    	JPanel buttonPanel;
    	JButton button;
    	JPanel centerPanel = new JPanel();;
        if (initialized)
            return;
        initialized = true;
        this.setSize(600, 200);
        Dimension windowSize = this.getSize();
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        this.setLocation(screenSize.width/2 - windowSize.width/2, screenSize.height/2 - windowSize.height/2);
        Container pane = this.getContentPane();
        pane.setLayout(new BorderLayout());
        centerPanel.setLayout(new BorderLayout());
        instructionText = new JTextArea();
        instructionText.setLineWrap(true);
        instructionText.setText(INSTRUCTIONS);
        instructionText.setEditable(false);
        instructionText.setBorder(new EmptyBorder(0,0,0,0));
        instructionText.setForeground(Color.BLACK);
        instructionText.setBackground(pane.getBackground());
        instructionText.setMinimumSize(new Dimension(100, 100));
        instructionText.setWrapStyleWord(true);
        centerPanel.add(new JScrollPane(instructionText), BorderLayout.NORTH);
        pane.add(centerPanel, BorderLayout.CENTER);
        statusText = new JTextField();
        statusText.setEditable(false);
        statusText.setBorder(new EmptyBorder(0,0,0,0));
        statusText.setForeground(Color.BLACK);
        statusText.setBackground(pane.getBackground());
        pane.add(new JScrollPane(statusText), BorderLayout.SOUTH);
        pane.add(mapFrame, BorderLayout.CENTER);
        buttonPanel = new JPanel();
        buttonPanel.setLayout(new BoxLayout(buttonPanel, BoxLayout.X_AXIS));
        
        button = new JButton("Clear Map");
        button.setActionCommand("clearmap");
        button.addActionListener(actions);
        button.addKeyListener(rovrKeyAdapter);
        buttonPanel.add(button);
        button = new JButton("Recenter Map");
        button.setActionCommand("recentermap");
        button.addActionListener(actions);
        button.addKeyListener(rovrKeyAdapter);
        buttonPanel.add(button);
        
        buttonPanel.add(new JLabel("Speed:"));
        buttonPanel.add(this.speedCombo);
        this.speedCombo.setSize(30, this.speedCombo.getHeight());

//		Leaning blows the automapper to hell        
        buttonPanel.add(new JLabel("Lean:"));
        buttonPanel.add(this.leanSpeedCombo);
        buttonPanel.add(this.leanDirectionCombo);
        this.leanDirectionCombo.setSize(30, this.leanDirectionCombo.getHeight());
        this.leanDirectionCombo.setSelectedIndex(1);
        this.leanSpeedCombo.setSize(30, this.leanSpeedCombo.getHeight());
        this.leanSpeedCombo.setSelectedIndex(0);

        buttonPanel.add(magnetCheckbox);
        buttonPanel.add(autopilotCheckbox);
        
        pane.add(buttonPanel, BorderLayout.NORTH);
    }
    
    /** 
     * Initialize the events.
     * It is very imporant that everything gets connected to the rovrKeyAdapter.  This means that
     * no matter what control is focused, the key adapter will receive the keypress events.
     * Also, this function takes care of subscribing everything to Client's movement event queue and
     * ties together the mapper and the map viewer.
     */
    private void initializeEvents() {
        this.client = new Client();
        client.addMovementListener(new MovementListenerStdout());
        client.addMovementListener(new MovementListenerTextField(statusText));
        client.addMovementListener(mapper);
        mapper.addMapChangeListener(mapFrame);
        client.addMovementListener(new StopListener());
        this.addKeyListener(rovrKeyAdapter);
        instructionText.addKeyListener(rovrKeyAdapter);
        statusText.addKeyListener(rovrKeyAdapter);
        speedCombo.addKeyListener(rovrKeyAdapter);
        leanDirectionCombo.addKeyListener(rovrKeyAdapter);
        leanSpeedCombo.addKeyListener(rovrKeyAdapter);
        magnetCheckbox.addKeyListener(rovrKeyAdapter);
        autopilotCheckbox.addKeyListener(rovrKeyAdapter);
        mapFrame.addKeyListener(rovrKeyAdapter);
        speedCombo.setActionCommand("speed");
        speedCombo.addActionListener(actions);
        leanSpeedCombo.setActionCommand("lean");
        leanSpeedCombo.addActionListener(actions);
        leanDirectionCombo.setActionCommand("lean");
        leanDirectionCombo.addActionListener(actions);
        magnetCheckbox.setSelected(true);
        autopilotCheckbox.setSelected(false);
        autopilotCheckbox.addActionListener(new AutopilotCheckListener());
        try {
        	client.connect();
        }catch(Throwable t) {
        	JOptionPane.showMessageDialog(this, "Could not connect to maze program", "Fatal Error", JOptionPane.ERROR_MESSAGE);
        	System.exit(0);
        }
    }
    
    /** Handles actions--events that occur when a user clicks on a widget. */
    public class Actions implements ActionListener {
		public void actionPerformed(ActionEvent e) {
            String command = e.getActionCommand();
            command = command == null ? "" : command;
            if (command.equals("showmap")) {
            	mapFrame.setVisible(true);
            	instructionText.grabFocus();
            } else if (command.equals("clearmap")) {
            	mapper.clear();
            } else if (command.equals("recentermap")) {
            	try {client.sendCommand(Client.COMMAND_STOP); }catch(Exception ex) {}
            	mapper.recenter(); 
            } else if (command.equals("speed")) {
            	client.setMoveFactor(speedCombo.getSelectedIndex() + 15);
	        } else if (command.equals("lean")) {
	        	client.setLean(leanDirectionCombo.getSelectedIndex() - 1, leanSpeedCombo.getSelectedIndex() +15);
	        }
        }
    }

    /** Die! */
    public void dispose() {
        // TODO: Save settings
        //super.dispose();
        System.exit(0);
    }
    
    /** Startup */
    public void setVisible(boolean b) {
        initialize();
        super.setVisible(b);
    }
    
    /**
     * Given a movement key, perform the desired action.
     * @param key
     */
    void movementKey(int key) {
    	try {
	    	switch(key) {
	    	case 'a':
	    		client.sendCommand(Client.COMMAND_UP);
	    		mapper.direction = Client.COMMAND_UP;
	    		break;
	    	case 'z':
	    		client.sendCommand(Client.COMMAND_DOWN);
	    		mapper.direction = Client.COMMAND_DOWN;
	    		break;
	    	case ',':
	    		client.sendCommand(Client.COMMAND_LEFT);
	    		mapper.direction = Client.COMMAND_LEFT;
	    		break;
	    	case '.':
	    		client.sendCommand(Client.COMMAND_RIGHT);
	    		mapper.direction = Client.COMMAND_RIGHT;
	    		break;
	    	case ' ':
	    		client.sendCommand(Client.COMMAND_STOP);
	    		mapper.direction = Client.COMMAND_STOP;
	    		break;
	    	case '1':
	    	case '2':
	    	case '3':
	    	case '4':
	    	case '5':
	    		speedCombo.setSelectedIndex(key - '1');
	    		client.setMoveFactor(speedCombo.getSelectedIndex() + 1);
	    		break;
	    	}
    	}catch(Exception e) {
    		System.out.println("Error sending command!");
    		// TODO: better handle exception
    	}
    }
    
    /** Program entry point */
	public static void main(String[] args) {
        new RovrFrame().setVisible(true);
	}
	
	/**
	 * Auto listener is an inner class that subscribed to Client's event
	 * queue.  If it gets a "I moved to another square" event, it checks
	 * the state of the magnet checkbox and issues a full stop if
	 * necessary.
     * 
     * 30 = hit wall going up/down
     * 40 = hit wall going left/right
	 */
	public class StopListener implements MovementListener {
		public void movementResponse(int code, String message) {
			try{
				if (magnetCheckbox.isSelected() && !autopilotCheckbox.isSelected()) {
					client.sendCommand(Client.COMMAND_STOP);
					mapper.direction = Client.COMMAND_STOP;
				}
			}catch(Exception e) {}
		}
	}

	public class AutopilotCheckListener implements ActionListener, AutopilotStatus {
        public void actionPerformed(ActionEvent e) {
        	System.out.println("Check listener fired");
        	if (autopilot != null)
        		autopilot.running = false;
        		autopilot = null;
            try {
                client.sendCommand(Client.COMMAND_STOP);
            }catch (Exception ex) {
                // TODO: what to do if exception?
            }
            if (autopilotCheckbox.isSelected()) {
                try {
                    speedCombo.setSelectedIndex(0);
                    leanSpeedCombo.setSelectedIndex(1);
                    leanDirectionCombo.setSelectedIndex(2);
                    magnetCheckbox.setSelected(false);
                    
                    speedCombo.setEnabled(false);
                    leanSpeedCombo.setEnabled(false);
                    leanDirectionCombo.setEnabled(false);
                    magnetCheckbox.setEnabled(false);

                    try{ Thread.sleep(1000); } catch(Throwable t) {}
                	autopilot = new CorridorAutopilot(this, client, mapper);
                	new Thread(autopilot).start();
                }catch (Exception ex) {
                    // TODO: what to do if exception?
                }
            }
        }
        public void autopilotFinished() {
            speedCombo.setEnabled(true);
            leanSpeedCombo.setEnabled(true);
            leanDirectionCombo.setEnabled(true);
            magnetCheckbox.setEnabled(true);
            autopilotCheckbox.setSelected(false);
        }
    }
	
	/**
	 * Handle key presses.  Mainly, this converts unconventiona;
	 * keys (function keys like the arrows) into ASCII keys. 
	 */
	public class RovrKeyAdapter extends KeyAdapter {
		public void keyPressed(KeyEvent e) {
			int rovrKey = 0;
			switch(e.getKeyChar()) {
			case 'A':
			case 'a':
				rovrKey = 'a';
				break;
			case 'Z':
			case 'z':
				rovrKey = 'z';
				break;
			case ',':
				rovrKey = ',';
				break;
			case '.':
				rovrKey = '.';
				break;
			case ' ':
				rovrKey = ' ';
				break;
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
				rovrKey = e.getKeyChar();
				break;
			}
			switch(e.getKeyCode()) {
			case KeyEvent.VK_KP_UP:
			case KeyEvent.VK_UP:
			case KeyEvent.VK_NUMPAD8:
				rovrKey = 'a';
				break;
			case KeyEvent.VK_KP_DOWN:
			case KeyEvent.VK_DOWN:
			case KeyEvent.VK_NUMPAD2:
				rovrKey = 'z';
				break;
			case KeyEvent.VK_KP_LEFT:
			case KeyEvent.VK_LEFT:
			case KeyEvent.VK_NUMPAD4:
				rovrKey = ',';
				break;
			case KeyEvent.VK_KP_RIGHT:
			case KeyEvent.VK_RIGHT:
			case KeyEvent.VK_NUMPAD7:
				rovrKey = '.';
				break;
			case KeyEvent.VK_NUMPAD5:
				rovrKey = ' ';
				break;
			}
			if (rovrKey != 0)
				movementKey(rovrKey);
		}
	}
}
